/*
 * Copyright 2005 - 2009  Zarafa B.V.
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License, version 3, 
 * as published by the Free Software Foundation with the following additional 
 * term according to sec. 7:
 *  
 * According to sec. 7 of the GNU Affero General Public License, version
 * 3, the terms of the AGPL are supplemented with the following terms:
 * 
 * "Zarafa" is a registered trademark of Zarafa B.V. The licensing of
 * the Program under the AGPL does not imply a trademark license.
 * Therefore any rights, title and interest in our trademarks remain
 * entirely with us.
 * 
 * However, if you propagate an unmodified version of the Program you are
 * allowed to use the term "Zarafa" to indicate that you distribute the
 * Program. Furthermore you may use our trademarks where it is necessary
 * to indicate the intended purpose of a product or service provided you
 * use it in accordance with honest practices in industrial or commercial
 * matters.  If you want to propagate modified versions of the Program
 * under the name "Zarafa" or "Zarafa Server", you may only do so if you
 * have a written permission by Zarafa B.V. (to acquire a permission
 * please contact Zarafa at trademark@zarafa.com).
 * 
 * The interactive user interface of the software displays an attribution
 * notice containing the term "Zarafa" and/or the logo of Zarafa.
 * Interactive user interfaces of unmodified and modified versions must
 * display Appropriate Legal Notices according to sec. 5 of the GNU
 * Affero General Public License, version 3, when you propagate
 * unmodified or modified versions of the Program. In accordance with
 * sec. 7 b) of the GNU Affero General Public License, version 3, these
 * Appropriate Legal Notices must retain the logo of Zarafa or display
 * the words "Initial Development by Zarafa" if the display of the logo
 * is not reasonably feasible for technical reasons. The use of the logo
 * of Zarafa in Legal Notices is allowed for unmodified and modified
 * versions of the software.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Affero General Public License for more details.
 *  
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 * 
 */

#include "platform.h"
#include "ECFifoBuffer.h"

#ifdef min
	#undef min	// We'll be using the stl version
#endif

#ifdef _DEBUG
#define new DEBUG_NEW
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;
#endif

class ECFifoBlock
{
public:
	ECFifoBlock(const void *lpBuf, unsigned int cbBuf);
	~ECFifoBlock();
	bool CopyTo(void *lpBuf, unsigned int ulOffset, unsigned int cbBuf, unsigned int *lpcbCopied);

private:
	char			*m_lpBuf;
	unsigned int	m_cbBuf;
	unsigned int	m_ulReadOffs;
};

ECFifoBlock::ECFifoBlock(const void *lpBuf, unsigned int cbBuf)
	: m_lpBuf(NULL)
	, m_cbBuf(cbBuf)
	, m_ulReadOffs(0)
{
	if (lpBuf && cbBuf > 0) {
		m_lpBuf = new char[cbBuf];
		memcpy(m_lpBuf, lpBuf, cbBuf);
	}
}

ECFifoBlock::~ECFifoBlock()
{
	if (m_lpBuf)
		delete[] m_lpBuf;
}

bool ECFifoBlock::CopyTo(void *lpBuf, unsigned int ulOffset, unsigned int cbBuf, unsigned int *lpcbCopied)
{
	unsigned int	cbNow = 0;

	if (lpBuf && cbBuf > 0 && ulOffset < cbBuf) {
		cbNow = std::min(cbBuf - ulOffset, m_cbBuf - m_ulReadOffs);
		memcpy((char*)lpBuf + ulOffset, m_lpBuf + m_ulReadOffs, cbNow);

		m_ulReadOffs += cbNow;
		if (lpcbCopied)
			*lpcbCopied = cbNow;

		if (m_ulReadOffs == m_cbBuf)
			return true;

	} else if (lpcbCopied)
		*lpcbCopied = 0;

	return false;
}


ECFifoBuffer::ECFifoBuffer()
	: m_ulMaxSize(8388608)
	, m_ulSize(0)
	, m_bClosed(false)
	, m_cbSWCache(0)
{ 
	pthread_mutex_init(&m_hMutex, NULL);
	pthread_cond_init(&m_hCond, NULL);
	
	m_lpszSWCache = new char[EC_FB_SWCACHE_SIZE];
}

ECFifoBuffer::ECFifoBuffer(unsigned int ulMaxSize)
	: m_ulMaxSize(ulMaxSize)
	, m_ulSize(0)
	, m_bClosed(false)
	, m_cbSWCache(0)
{ 
	pthread_mutex_init(&m_hMutex, NULL);
	pthread_cond_init(&m_hCond, NULL);

	m_lpszSWCache = new char[EC_FB_SWCACHE_SIZE];
}

ECFifoBuffer::~ECFifoBuffer()
{
	pthread_mutex_destroy(&m_hMutex);
	pthread_cond_destroy(&m_hCond);
	
	for (BlockIter block = m_sBlockList.begin(); block != m_sBlockList.end(); ++block)
		delete *block;
	
	delete[] m_lpszSWCache;
}
	
ECRESULT ECFifoBuffer::Write(const void *lpBuf, unsigned int cbBuf, unsigned int ulTimeoutMs)
{
	ECRESULT		er = erSuccess;
	struct timeval	now = {0};
	struct timespec	timeout = {0};

	if (lpBuf == NULL)
		return ZARAFA_E_INVALID_PARAMETER;

	if (cbBuf == 0)
		return erSuccess;

	if (cbBuf > m_ulMaxSize)
		return ZARAFA_E_NOT_ENOUGH_MEMORY;

	if (ulTimeoutMs > 0) {
		gettimeofday(&now, NULL);

		now.tv_sec += ulTimeoutMs / 1000;
		now.tv_usec += 1000 * (ulTimeoutMs % 1000);
		if (now.tv_usec >= 1000000) {
			now.tv_sec++;
			now.tv_usec -= 1000000;
		}

		timeout.tv_sec = now.tv_sec;
		timeout.tv_nsec = now.tv_usec * 1000;
	}

	pthread_mutex_lock(&m_hMutex);

	if (m_bClosed) {
		er = ZARAFA_E_CALL_FAILED;
		goto exit;
	}

	// First wait until we wouldn't go boyond the maximum size
	while (m_ulSize + cbBuf > m_ulMaxSize) {
		if (ulTimeoutMs > 0) {
			if (pthread_cond_timedwait(&m_hCond, &m_hMutex, &timeout) == ETIMEDOUT) {
				er = ZARAFA_E_TIMEOUT;
				goto exit;
			}
		} else
			pthread_cond_wait(&m_hCond, &m_hMutex);
	}

	// If the current amount of data would overflow the SWCache, flush it first
	if (m_cbSWCache + cbBuf > EC_FB_SWCACHE_SIZE) {
		m_sBlockList.push_back(new ECFifoBlock(m_lpszSWCache, m_cbSWCache));
		m_cbSWCache = 0;
	}

	// Check if the data to write would fit in the SWCache. It could have been
	// just flushed but just too large.
	if (m_cbSWCache + cbBuf > EC_FB_SWCACHE_SIZE)
		m_sBlockList.push_back(new ECFifoBlock(lpBuf, cbBuf));
	else {
		memcpy(m_lpszSWCache + m_cbSWCache, lpBuf, cbBuf);
		m_cbSWCache += cbBuf;
	}

	m_ulSize += cbBuf;

exit:
	pthread_cond_signal(&m_hCond);
	pthread_mutex_unlock(&m_hMutex);

	return er;
}

ECRESULT ECFifoBuffer::Read(void *lpBuf, unsigned int cbBuf, unsigned int ulTimeoutMs, unsigned int *lpcbRead)
{
	ECRESULT		er = erSuccess;
	unsigned int	cbRead = 0;
	unsigned int	cbNow = 0;
	struct timeval	now = {0};
	struct timespec	timeout = {0};

	if (lpBuf == NULL)
		return ZARAFA_E_INVALID_PARAMETER;

	if (cbBuf == 0) {
		if (lpcbRead != NULL)
			*lpcbRead = 0;
		return erSuccess;
	}

	if (ulTimeoutMs > 0) {
		gettimeofday(&now, NULL);

		now.tv_sec += ulTimeoutMs / 1000;
		now.tv_usec += 1000 * (ulTimeoutMs % 1000);
		if (now.tv_usec >= 1000000) {
			now.tv_sec++;
			now.tv_usec -= 1000000;
		}

		timeout.tv_sec = now.tv_sec;
		timeout.tv_nsec = now.tv_usec * 1000;
	}

	pthread_mutex_lock(&m_hMutex);

	while (cbRead < cbBuf) {
		if (m_sBlockList.empty()) {
			if (m_bClosed)
				break;
			if (ulTimeoutMs > 0) {
				if (pthread_cond_timedwait(&m_hCond, &m_hMutex, &timeout) == ETIMEDOUT) {
					er = ZARAFA_E_TIMEOUT;
					goto exit;
				}
			} else
				pthread_cond_wait(&m_hCond, &m_hMutex);
		} else {
			if (m_sBlockList.front()->CopyTo(lpBuf, cbRead, cbBuf, &cbNow)) {
				delete m_sBlockList.front();
				m_sBlockList.pop_front();
			}
			cbRead += cbNow;
			pthread_cond_signal(&m_hCond);
		}
	}

exit:
	m_ulSize -= cbRead;
	if (lpcbRead)
		*lpcbRead = cbRead;

	pthread_cond_signal(&m_hCond);
	pthread_mutex_unlock(&m_hMutex);

	return er;
}


ECRESULT ECFifoBuffer::Close()
{
	pthread_mutex_lock(&m_hMutex);
	m_bClosed = true;
	// Flush the SWCache
	if (m_cbSWCache > 0) {
		m_sBlockList.push_back(new ECFifoBlock(m_lpszSWCache, m_cbSWCache));
		m_cbSWCache = 0;
	}
	pthread_cond_signal(&m_hCond);
	pthread_mutex_unlock(&m_hMutex);
	return erSuccess;
}
